from .ubki_credit_report import get_useful_credit_report_fields
from .ubki_credit_score import get_useful_credit_score_fields

from datetime import datetime
from typing import List

import xmltodict

class UbkiReport():
    """ Класс отчет УБКИ

    Parameters
    ----------
    xml_credit_report : str
        Кредитный отчет физической особы, предпринимателя
        
    xml_credit_score : str
        Кредитный балл 

    phone : str
        Телефон 

    email : str
        Электронная почта
    """
    def __init__(self, xml_credit_report:str, xml_credit_score:str, phone:str, email:str):
        self.xml = {'report': xml_credit_report, 'score': xml_credit_score}
        self.phone = phone
        self.email = email

    def get_report_xml(self) -> str:
        """ Метод получения кредитного отчета физической особы, предпринимателя

        Returns
        -------
        ubki report : Кредитный отчет физической особы, предпринимателя
        """
        return self.xml['report']
    def get_score_xml(self) -> str:
        """ Метод получения кредитный балл

        Returns
        -------
        ubki report : Кредитний бал
        """
        return self.xml['score']

    def get_useful_ubki_fields(self, fields_to_ignore: List[str] = []) -> dict:
        """ Метод получения полезных полей из УБКИ отчета

        Parameters
        ----------
        fields_to_ignore : List[str] = []
            Список не нужных параметров, который следует исключить из итогового набора

        Returns
        -------
        useful ubki fields : Словарь полезных параметром для скоринга
        """
        ubki = {
                **get_useful_credit_report_fields(xmltodict.parse(self.xml['report']), self.phone, self.email, datetime.now()),
                **get_useful_credit_score_fields(xmltodict.parse(self.xml['score']))
                }
        return {key: ubki[key] for key in ubki if not key in fields_to_ignore}


def get_useful_credit_score_fields(ubki: dict) -> dict:
    """ Функция получения полезных параметров из кредитного балла

    Parameters
    ----------
    ubki : dict
        УБКИ отчет в виде словаря

    Returns
    -------
    useful ubki fields : Словарь полезных параметром для скоринга из кредитного балла
    """
    res_dict = {
        k: None
        for k in [
            "ubki_score",           # УБКИ очки
            "ubki_scorelast",       # УБКИ последние очки
            "ubki_scorelevel",      # УБКИ уровень очков
            "ubki_all_credits",     # Все кредиты
            "ubki_open_credits",    # Открытые кредиты
            "ubki_closed_credits",  # Закрытые кредиты
            "ubki_expyear",         # Просрочка по кредитам
            "ubki_maxnowexp",       #
            "ubki_phone_deltatime", # Разница между первым упоминанием данного телефена и текущей датой
            "ubki_email_deltatime", # Разница между первым упоминанием данного почты и текущей датой
            "ubki_week_queries"     # Количество запросов в неделю
        ]
    }
    # print(ubki)
    if not ubki:
        return res_dict
    try:
        if 'ubkidata' in ubki:
            ubki_response = ubki['ubkidata']

        if "comp" in ubki_response.keys():
            for comp in ubki_response["comp"]:
                comp_id = comp["@id"]
    
                if comp_id == "8": # Кредитный рейтинг УБКИ
                    try:
                        rating = comp["urating"]
                        for i in ["@score", "@scorelast", "@scorelevel"]:
                            res_dict["ubki_" + i[1:]] = int(float(rating[i])) \
                                if (check_null_value(rating[i])) else None

                        dinfo = rating["dinfo"]
                        res_dict["ubki_all_credits"] = int(float(dinfo["@all"])) \
                            if (check_null_value(rating[i])) else None
                        res_dict["ubki_open_credits"] = int(float(dinfo["@open"])) \
                            if (check_null_value(dinfo["@open"])) else None
                        res_dict["ubki_closed_credits"] = int(float(dinfo["@close"])) \
                            if (check_null_value(dinfo["@close"])) else None
                        res_dict["ubki_expyear"] = int(coding_no_yes(dinfo["@expyear"])) \
                            if (check_null_value(dinfo["@expyear"])) else None
                        res_dict["ubki_maxnowexp"] = int(float(coding_maxnowexp(dinfo["@maxnowexp"]))) \
                            if (check_null_value(dinfo["@maxnowexp"])) else None
                    finally:
                        break

    except:
        pass
    finally:
        return res_dict

def check_null_value(value: str) -> bool:
    return value != '' and value != "NA" and value != [] and value != "null"


def coding_no_yes(arg: str) -> int:
    return 0 if (arg.lower() == "нет") else 1


def coding_maxnowexp(arg: str) -> int:
    return 0 if (arg.lower() == "нет") else arg.split(" ")[0]


from dotenv import dotenv_values, load_dotenv, set_key, find_dotenv
from .ubki_report import UbkiReport

from datetime import datetime
from ast import literal_eval
from typing import Optional
import requests
import json

class UbkiRequest ():
    """ Класс инициализатор подключения к УБКИ.
    Предоставляет авторизацию и возможность посылания запросов на УБКИ, с целью
    получения кредитной истории конкретной персоны.

    Parameters
    ----------
    login : str
        Логин партнера для авторизации
        
    password : str
        Пароль партнера для авторизации
    
    is_test : bool = False
        Логическое выражения для опрделения запросов на тестовый сервер или оригинальный
    """
    def __init__ (self, login:str, password:str, is_test:Optional[bool] = False):
        self.is_test = is_test
        if is_test:
            self.ubki_url = "https://test.ubki.ua/b2_api_xml/ubki/xml"
        else: 
            self.ubki_url = "https://secure.ubki.ua/b2_api_xml/ubki/xml"

        config, dotenv_file = self._env()
        pref = "test" if is_test else "real"

        last_key = datetime.strptime(config[pref + '_last_key'],'%Y-%m-%d')
        now = datetime.now().date()

        if last_key.date() == now:
            self.sessid = config[pref + '_key']
        else:
            self.sessid = self.ubki_authorization(login, password)
            set_key(dotenv_file, pref + "_last_key", now.strftime('%Y-%m-%d'))
            set_key(dotenv_file, pref + "_key", self.sessid)
            print("Успешная Авторизация!!!")

    def ubki_authorization(self, login: str, password: str) -> str:
        """ Метод авторизации и получения сессионого ключа

        Parameters
        ----------
        login : str
            Логин партнера для авторизации
        
        password : str
            Пароль партнера для авторизации

        Returns
        -------
        sessid : Сессионный ключ
        """
        if self.is_test:
            url = "https://test.ubki.ua/b2_api_xml/ubki/auth"
        else:
            url = "https://secure.ubki.ua/b2_api_xml/ubki/auth"
        request_text = json.dumps(
        {
            "doc": {
                "auth": {
                    "login": login,
                    "pass": password
                }
            }
        })
        headers = {'Content-Type': 'application/json', 'Accept': 'application/json'}
        response = requests.request("POST", url, headers=headers, data=request_text)
        return literal_eval(response.text)['doc']['auth']['sessid']

    def get_person_credit_report(self, person_data:Optional[dict] = None) -> UbkiReport:
        """ Метод получения отчета убки об интересующей персоне

        Parameters
        ----------
        person_data : dict = None
            Словарь с необходимыми данными о искомой персоне

        Returns
        -------
        ubki report : Отчет с получеными данными о искомой персоне
        """
        if person_data != None and not self.is_test:
            data = person_data
        elif person_data == None and self.is_test:
            data = \
            {"okpo":"2111118724", 
            "lname":"РИБАЧКА", "fname":"АННА", "mname":"ІГОРЕВНА", "bdate":"1957-10-19", 
            "dtype":"1", "dser":"ВВ", "dnom":"142228", 
            "cval":"+380111656411", "ctype":"3", 'email':"email@gmail.com", 'reqidout':'00001'}
        else:
            raise Exception("ERROR, mode mismatch!!")
        return UbkiReport(self._request(10, data), self._request(11, data), data['cval'], data['email'])

    def _request(self, reqtype:int, data:dict) -> str:
        """ Метод отправки запроса на получения убки отчета о искомой персоне

        Parameters
        ----------
        reqtype : int
            Номер шаблона отчета убки, который требуется получить

        data : dict
            Данные о искомой персоне, необходимые для запроса

        Returns
        -------
        response : Отчет убки полученный по запросу на искомою персону
        """
        request_text =  "<?xml version=\"1.0\" encoding=\"utf-8\" ?>"\
        "<doc>"\
            f"<ubki sessid=\"{self.sessid}\">"\
                "<req_envelope descr=\"Конверт запиту\">"\
                    "<req_xml descr=\"Об'єкт запиту\">"\
                         f"<request version=\"1.0\" reqtype=\"{reqtype}\" reqreason=\"2\" "\
                                                   f"reqdate=\"{datetime.now().strftime('%Y-%m-%d')}\" "\
                                                   f"reqidout=\"{data['reqidout']}\" reqsource=\"1\">"\
                            "<i reqlng=\"1\">"\
                                f"<ident okpo=\"{data['okpo']}\" lname=\"{data['lname']}\" "\
                                    f"fname=\"{data['fname']}\" mname=\"{data['mname']}\" bdate=\"{data['bdate']}\" />"\
                                f"<spd inn=\"{data['okpo']}\" />"\
                                "<docs>"\
                                    f"<doc dtype=\"{data['dtype']}\" dser=\"{data['dser']}\" dnom=\"{data['dnom']}\" />"\
                                "</docs>"\
                                "<contacts>"\
                                    f"<cont cval=\"{data['cval']}\" ctype=\"{data['ctype']}\" />"\
                                "</contacts>"\
                            "</i>"\
                        "</request>"\
                    "</req_xml>"\
                "</req_envelope>"\
            "</ubki>"\
        "</doc>"
        return requests.request("POST", self.ubki_url, data=request_text.encode('utf-8')).text

    def _env(self):
        """ 
        Метод подключения к .env и получения необходимых переменных среды, для работы
        """
        dotenv_file = find_dotenv()
        if dotenv_file == '': # Создание .env файла в случае отсуствия оного
                with open('.env', 'w') as file:
                    file.write("test_key=1\nreal_key=1\n")
                    file.write('test_last_key=2000-1-1\nreal_last_key=2000-1-1')
                dotenv_file = find_dotenv()
        load_dotenv(dotenv_file)
        return dotenv_values(), dotenv_file


from typing import List
import datetime as dt
import numpy as np
import datetime

def get_useful_credit_report_fields(ubki: dict, phone: str, email: str, our_date: datetime) -> dict:
    """ Функция получения полезных параметров из кредитного отчета физического лица

    Parameters
    ----------
    ubki : dict
        УБКИ отчет в виде словаря

    phone : str
        Телефон, для вычисления самой раннего упоминания даного контакта

    email : str
        Электронная почта, для вычисления самой раннего упоминания даного контакта

    our_date : datetime
        Дата получения отчета

    Returns
    -------
    useful ubki fields : Словарь полезных параметром для скоринга из кредитного отчета
    """
    res_dict = {
        k: None
        for k in [
            "median_day_credit",    # Медианна по дате начала соглашения
            "mean_credit_summ",     # Средняя сумма по кредитам
            "mean_credit_debt",     # Средння задолжность по кредитам
            "cdolgn",               # Последняя зафиксированая должность на работе
            "wdohod",               # Последний зафиксированый доход на работе
            "wstag",                # Последний зафиксированый стаж на работе
            "max_wdohod",           # Максимальный зафиксировынный доход
            "cgrag",                # Код странны (категориальный)
            "sstate",               # Социальный статус (категориальный)
            "family",               # Симейный статус (категориальный)
            "ceduc",                # Образование (категориальный)
            "ubki_balance_value",   # Текущий баланс
            "ubki_phone_deltatime", # Разница между первым упоминанием данного телефена и текущей датой
            "ubki_email_deltatime", # Разница между первым упоминанием данного почты и текущей датой
            "ubki_week_queries"     # Количество запросов в неделю
        ]
    }
    try:
        if 'ubkidata' in ubki:
            ubki_response = ubki['ubkidata']
        try: # Системные данные
            if "tech" in ubki_response.keys() and "billing" in ubki_response['tech'].keys(): 
                res_dict["ubki_balance_value"] = int(
                    float(ubki_response['tech']["billing"]["balance"]["@value"]))  # Баланс 
        except:
            pass
        # Персональные данные
        if "comp" in ubki_response.keys():
            for comp in ubki_response["comp"]:
                comp_id = comp["@id"]
                
                if comp_id == '1': # Блок персональных данных

                    try: # Данные про субьект
                        comp_ident = comp['cki']['ident']
                        if type(comp_ident) == list:
                            for ident in comp_ident:
                                fill_dict_by_key(ident, ["@cgrag", '@sstate', '@family', '@ceduc'], res_dict)
                        else:
                            fill_dict_by_key(ident, ["@cgrag", '@sstate', '@family', '@ceduc'], res_dict)
                    except:
                        pass
                    
                    try: # Данные про роботу
                        comp_work = comp['cki']['work']
                        if type(comp_work) == list:
                            for work in comp_work:
                                fill_dict_by_key(work, ['@cdolgn', '@wdohod', '@wstag'], res_dict)
                        else:
                            fill_dict_by_key(comp_work, ['@cdolgn', '@wdohod', '@wstag'], res_dict)
                    except :
                        pass

                elif comp_id == '2': # Блок информации про кредитные соглашения
                    try:
                        ubki_crdeal = comp['crdeal']
                        credit_sum = credits = 0        # Сумма кредита и количество кредитов
                        credit_debt = debts = 0         # Сумма задолжности по кредиту
                        median_days = np.array([0, 0])  # Сумма по дням в половинах месяца
                        if type(ubki_crdeal) == list:   # Проверка на количество кредитов
                            for credit in ubki_crdeal:
                                amount, debt, days = get_credit_ubki_fields(credit)
                                if amount > 0:          # Проверка на наличие кредитной суммы
                                    credit_sum += amount
                                    credits += 1
                                if debt > 0:            # Проверка на наличие задолжности по кредитному соглашению
                                    credit_debt += debt
                                    debts += 1
                                median_days += days
                        else:                           # Только один кредит
                            credit_sum, credit_debt, median_days = get_credit_ubki_fields(ubki_crdeal)
                            credits = 1 if (credit_sum > 0) else 0
                            debts = 1 if (credit_debt > 0) else 0

                        res_dict['median_day_credit'] = int(np.argmax(median_days))
                        res_dict['mean_credit_summ'] = int(credit_sum / credits)
                        res_dict['mean_credit_debt'] = int(credit_debt / debts)
                    except:
                        pass

                elif comp_id == "4": # Блок регестрации запросов
                    try:
                        res_dict["ubki_week_queries"] = int(comp["reestrtime"]["@wk"])  # Количество запросов в неделю
                        res_dict["req_credit"] = sum([1 for i in comp['credres'] if i['@reqreason'] in ['2', '4']]) # Количество запросов на кредит
                    except:
                        pass

                if comp_id == "10": # Блок истории контактных данных
                    try:
                        big_datetime = dt.datetime(dt.datetime.today().year + 1, 1, 1)
                        first_datetime_phone = first_datetime_email = big_datetime
                        for contact in comp["cont"]:
                            cval = contact["@cval"]
                            vdate = contact["@vdate"]

                            # Нахождения самого раннего упоминания телефона
                            if cval.replace('+', '') == str(phone).replace('+', ''):
                                if format_date(vdate) < first_datetime_phone:
                                    first_datetime_phone = format_date(vdate)

                            # Нахождения самого раннего упоминания електронной почты
                            elif str(cval.lower()) == str(email):
                                if format_date(vdate) < first_datetime_email:
                                    first_datetime_email = format_date(vdate)

                        res_dict["ubki_phone_deltatime"] = int((our_date - first_datetime_phone).days) if (
                                first_datetime_phone != big_datetime) else None
                        res_dict["ubki_email_deltatime"] = int((our_date - first_datetime_email).days) if (
                                first_datetime_email != big_datetime) else None
                    except:
                        pass
    except:
        print("Ошибка УБКИ")
    finally:
        return res_dict

def fill_dict_by_key(values:dict, keys:List[str], full_dict:dict) -> dict:
    """ Функция получения списка значения по списку ключей

    Parameters
    ----------
    values : dict
        Словарь с данными

    keys : List[str] 
        Список необходимых ключей

    full_dict : dict
        Полный словарь, к которому добавяться данные

    Returns
    -------
    full dict: Словарь с полученными всеми или присутствубщимы данными
    """
    for key in keys:
        full_dict[key[1:]] = float(values[key]) if check_null_value(values[key]) else full_dict[key[1:]]
    return full_dict

def get_credit_ubki_fields(credit: dict) -> float:
    """ Функция получения полезной информации о каждом кредите

    Parameters
    ----------
    credit : dict
        Кредит, который будет источником информации

    Returns
    -------
    userful credit fields : Полезные значения для скоринга о кредитных соглашениях"""
    credit_amount = float(credit['@dlamt']) # Сумма кредита
    credit = credit['deallife']

    # Проверка на количество записей в кредитной истории
    if type(credit) == list:
        credit_date = format_date(credit[0]['@dlds'])
        credit_debt = max([float(dl['@dlamtcur']) for dl in credit]) # Вычесление максимальной задолжности
    else:
        credit_date = format_date(credit['@dlds'])
        credit_debt = float(credit['@dlamtcur'])
    median_days =  [1, 0] if credit_date.day < 15 else [0, 1]
    
    return credit_amount, credit_debt, median_days

def format_date(attribute: str) -> datetime:
    return dt.datetime.strptime(attribute, "%Y-%m-%d")

def check_null_value(value: str) -> bool:
    return value != '' and value != "NA" and value != [] and value != "null"


from setuptools import find_packages, setup

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="ubkisaas",
    version="0.1",
    author="Asir Muminov",
    author_email="vojt.tieg295i@gmail.com",
    description="library for receiving UBKI data from the official website and parsing them and getting features for scoring analysis",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Asirg/ubkisaas",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    # package_dir={"": "ubki"},
    packages=find_packages(),
    python_requires=">=3.6",
    install_requires=[
        "xmltodict",
        "requests",
        "typing",
        "python-dotenv",
        "numpy"
    ]
)